CREATE OR REPLACE PACKAGE Saper AS

  --  
  PROCEDURE print(message IN VARCHAR2); 
  --  
  FUNCTION get_player(name IN saper_players.name%TYPE)
    RETURN saper_players%ROWTYPE;
  --  
  PROCEDURE add_player(name IN saper_players.name%TYPE);
  --  
  FUNCTION get_level(name IN saper_levels.name%TYPE)
    RETURN saper_levels%ROWTYPE;
  --  
  PROCEDURE add_level(name IN saper_levels.name%TYPE,
    width IN saper_levels.width%TYPE, height IN saper_levels.height%TYPE,
    mines IN saper_levels.mines%TYPE);
  --  
  PROCEDURE add_game(level IN saper_levels.name%TYPE, player IN saper_players.name%TYPE);
  --  
  PROCEDURE delete_game(del_id IN saper_games.id%TYPE);
  --  
  PROCEDURE print_game(game_id IN saper_games.id%TYPE);
  --  
  PROCEDURE mark(game_id IN saper_cells.game_id%TYPE,
    cell_x IN saper_cells.x%TYPE, cell_y IN saper_cells.y%TYPE);
  --  
  PROCEDURE look(game_id IN saper_cells.game_id%TYPE,
    cell_x IN saper_cells.x%TYPE, cell_y IN saper_cells.y%TYPE);
  --   ( 3x3)
  PROCEDURE near(game_id IN saper_cells.game_id%TYPE,
    cell_x IN saper_cells.x%TYPE, cell_y IN saper_cells.y%TYPE);

END saper;
/
CREATE OR REPLACE PACKAGE BODY Saper AS


  --  
  PROCEDURE print(message IN VARCHAR2)
  IS
  BEGIN
    dbms_output.enable;
    dbms_output.put_line(message);
  END print;
  
  --  
  FUNCTION get_player(name IN saper_players.name%TYPE)
    RETURN saper_players%ROWTYPE
  IS
    CURSOR player_cur(name_in IN saper_players.name%TYPE)
      RETURN saper_players%ROWTYPE
    IS
      SELECT * FROM saper_players WHERE name=name_in;
    player saper_players%ROWTYPE;
  BEGIN
    OPEN player_cur(name);
    FETCH player_cur INTO player;
    IF player_cur%NOTFOUND THEN
      RETURN NULL;
    END IF;
    CLOSE player_cur;
    RETURN player;
  END get_player;
  
  --  
  PROCEDURE add_player(name IN saper_players.name%TYPE)
  IS
  BEGIN
    IF get_player(name).id IS NULL THEN
      INSERT INTO saper_players VALUES(saper_players_seq.NEXTVAL, name);
      COMMIT;
    END IF;
  END add_player;
  
  --  
  FUNCTION get_level(name IN saper_levels.name%TYPE)
    RETURN saper_levels%ROWTYPE
  IS
    CURSOR level_cur(name_in IN saper_levels.name%TYPE)
      RETURN saper_levels%ROWTYPE
    IS
      SELECT * FROM saper_levels WHERE name=name_in;
    level saper_levels%ROWTYPE;
  BEGIN
    OPEN level_cur(name);
    FETCH level_cur INTO level;
    IF level_cur%NOTFOUND THEN
      RETURN NULL;
    END IF;
    CLOSE level_cur;
    RETURN level;
  END get_level;
    
  --  
  PROCEDURE add_level(name IN saper_levels.name%TYPE,
    width IN saper_levels.width%TYPE, height IN saper_levels.height%TYPE,
    mines IN saper_levels.mines%TYPE)
  IS
  BEGIN
    IF get_level(name).id IS NULL THEN
      INSERT INTO saper_levels VALUES(saper_levels_seq.NEXTVAL,
        name, width, height, mines);
      COMMIT;
    END IF;
  END add_level;
  
  --  
  FUNCTION get_cell(x IN saper_cells.x%TYPE, y IN saper_cells.y%TYPE,
    game IN saper_cells.game_id%TYPE)
    RETURN saper_cells%ROWTYPE
  IS
    CURSOR cell_cur(x_in IN saper_cells.x%TYPE, y_in IN saper_cells.y%TYPE,
      game_in IN saper_cells.game_id%TYPE)
      RETURN saper_cells%ROWTYPE
    IS
      SELECT * FROM saper_cells WHERE x=x_in AND y=y_in AND game_id=game_in;
    cell saper_cells%ROWTYPE;
  BEGIN
    OPEN cell_cur(x,y,game);
    FETCH cell_cur INTO cell;
    IF cell_cur%NOTFOUND THEN
      RETURN NULL;
    END IF;
    CLOSE cell_cur;
    RETURN cell;
  END get_cell;
  
  --  
  PROCEDURE put_mine(game IN saper_cells.game_id%TYPE,
    width IN saper_levels.width%TYPE, height IN saper_levels.height%TYPE)
  IS
    mine_x saper_cells.x%TYPE;
    mine_y saper_cells.y%TYPE;
    mine saper_cells%ROWTYPE;
  BEGIN
    --  ,   
    LOOP
      SELECT MOD(ABS(dbms_random.random),width) INTO mine_x FROM DUAL;
      SELECT MOD(ABS(dbms_random.random),height) INTO mine_y FROM DUAL;
      mine:=get_cell(mine_x,mine_y,game);
      EXIT WHEN mine.mine=0;
    END LOOP;
    --     
    UPDATE saper_cells
      SET mine=1
      WHERE id=mine.id;
    COMMIT;
  END put_mine;
  
  --  
  FUNCTION get_game(game_id IN saper_games.id%TYPE)
    RETURN saper_games%ROWTYPE
  IS
    CURSOR game_cur(game_id IN saper_games.id% TYPE)
      RETURN saper_games%ROWTYPE
    IS
      SELECT * FROM saper_games WHERE id=game_id;
    game saper_games%ROWTYPE;
  BEGIN
    OPEN game_cur(game_id);
    FETCH game_cur INTO game;
    IF game_cur%NOTFOUND THEN
      RETURN NULL;
    END IF;
    CLOSE game_cur;
    RETURN game;
  END get_game;
  
  --  
  PROCEDURE create_game(level_id IN saper_games.level_id%TYPE,
    width IN saper_levels.width%TYPE, height IN saper_levels.height%TYPE,
    player IN saper_players.name%TYPE, mines IN saper_levels.mines%TYPE)
  IS
    player_id saper_games.player_id%TYPE;
    new_id saper_games.id%TYPE;
    near_count saper_cells.near%TYPE;
  BEGIN
    player_id:=get_player(player).id;
    IF (NOT(player_id IS NULL)) THEN
      --  
      SELECT saper_games_seq.NEXTVAL INTO new_id FROM DUAL;
      INSERT INTO saper_games VALUES (new_id,level_id, player_id, SYSDATE);
      --   
      FOR x IN 0..width-1
      LOOP
        FOR y IN 0..height-1
        LOOP
          INSERT INTO saper_cells (id,x,y,game_id,near,mine,state) VALUES
            (saper_cells_seq.NEXTVAL, x, y,new_id, 0, 0, 0);
        END LOOP;
      END LOOP;
      --    
      dbms_random.seed(to_char(SYSDATE,'MM-DD-YYYY HH24:MI:SS'));
      FOR i IN 0..mines-1
      LOOP
        put_mine(new_id,width,height);
      END LOOP;
      --   
      FOR cell_x IN 0..width-1
      LOOP
        FOR cell_y IN 0..height-1
        LOOP
          SELECT
            (SELECT COUNT(*) FROM saper_cells
            WHERE ABS(x-cell_x)<=1 AND ABS(y-cell_y)<=1
            AND game_id=new_id AND mine=1)
            INTO near_count FROM DUAL;
          UPDATE saper_cells
            SET near=near_count
            WHERE game_id=new_id and x=cell_x and y=cell_y;
        END LOOP;
      END LOOP;
      --  
      COMMIT;
      print('ID  : '||to_char(new_id));
      print_game(new_id);
    END IF;
  END create_game;

  --  
  PROCEDURE add_game(level IN saper_levels.name%TYPE, player IN saper_players.name%TYPE)
  IS
    lev saper_levels%ROWTYPE;
  BEGIN
    lev:=get_level(level);
    IF NOT(lev.id IS NULL) THEN
      create_game(lev.id,lev.width,lev.height,player,lev.mines);
    END IF;
  END add_game;
  
  --  
  PROCEDURE delete_game(del_id IN saper_games.id%TYPE)
  IS
  BEGIN
    --   
    DELETE FROM saper_cells
      WHERE game_id=del_id;
    -- 
    DELETE FROM saper_games
      WHERE id=del_id;
  END delete_game;
  
  --    
  FUNCTION print_cell_clear(cell IN saper_cells%ROWTYPE)
    RETURN VARCHAR
  IS
    sym VARCHAR(2);
  BEGIN
    IF cell.mine=0 THEN
      IF cell.near=0 THEN
        sym:=' ';
      ELSE
        sym:=to_char(cell.near);
      END IF;
    ELSE
      IF cell.state=2 THEN
        sym:='P';
      ELSE
        sym:='*';
      END IF;
    END IF;
    RETURN sym;
  END print_cell_clear;
  
  --   
  FUNCTION print_cell_visible(cell IN saper_cells%ROWTYPE)
    RETURN VARCHAR
  IS
    sym VARCHAR(2);
  BEGIN
    IF cell.state=0 THEN
      sym:='#';
    ELSIF cell.state=2 THEN
      sym:='P';
    ELSIF cell.near=0 THEN
      sym:=' ';
    ELSE
      sym:=to_char(cell.near);
    END IF;
    RETURN sym;
  END print_cell_visible;
  
  --   
  PROCEDURE print_in_some_way(game_id IN saper_games.id%TYPE, clearly IN NUMBER)
  IS
    game saper_games%ROWTYPE;
    output VARCHAR2(100);
    sym VARCHAR2(2);
    cell saper_cells%ROWTYPE;
    height saper_levels.height%TYPE;
    width saper_levels.width%TYPE;
  BEGIN
    game:=get_game(game_id);
    IF NOT(game.id IS NULL) THEN
    SELECT L.width,L.height INTO width, height FROM saper_levels L WHERE id=game.level_id;
    --   
    output:='.';
    FOR cell_x IN 0..width-1
    LOOP
      sym:=LPAD(to_char(cell_x),2,'_');
      output:=output||sym;
    END LOOP;
    output:=output||'.';
    print(output);
    --  
      FOR cell_y IN 0..height-1
      LOOP
        output:='|';
        FOR cell_x IN 0..width-1
        LOOP
          cell:=get_cell(cell_x,cell_y,game_id);
          IF clearly=1 THEN
            sym:=print_cell_clear(cell);
          ELSE
            sym:=print_cell_visible(cell);
          END IF;
          output:=output||LPAD(sym,2);
        END LOOP;
        output:=output||'|'||LPAD(to_char(cell_y),2);
        print(output);
      END LOOP;
    END IF;
  END print_in_some_way;
  
  --   
  PROCEDURE print_clear(game_id IN saper_games.id%TYPE)
  IS
  BEGIN
    print_in_some_way(game_id,1);
  END print_clear;
  
  --  
  PROCEDURE print_game(game_id IN saper_games.id%TYPE)
  IS
  BEGIN
    print_in_some_way(game_id,0);
  END print_game;
  
  --  
  PROCEDURE mark(game_id IN saper_cells.game_id%TYPE,
    cell_x IN saper_cells.x%TYPE, cell_y IN saper_cells.y%TYPE)
  IS
    cell saper_cells%ROWTYPE;
  BEGIN
    cell:=get_cell(cell_x,cell_y,game_id);
    IF NOT(cell.id IS NULL) THEN
      IF cell.state=0 OR cell.state=2 THEN
        UPDATE saper_cells
          SET state=2-cell.state
          WHERE id=cell.id;
        COMMIT;
      END IF;
    END IF;
    print_game(game_id);
  END mark;
  
  --  
  PROCEDURE look(game_id IN saper_cells.game_id%TYPE,
    cell_x IN saper_cells.x%TYPE, cell_y IN saper_cells.y%TYPE)
  IS
    cell saper_cells%ROWTYPE;
  BEGIN
    cell:=get_cell(cell_x,cell_y,game_id);
    IF NOT(cell.id IS NULL) THEN
      IF cell.state=0 THEN
        UPDATE saper_cells
          SET state=1
          WHERE id=cell.id;
        COMMIT;
      END IF;
    END IF;
    IF cell.mine=1 THEN
      print('YOU LOSE!');
      print_clear(game_id);
      delete_game(game_id);
    ELSE
      print_game(game_id);
    END IF;
  END look;
  
  --   ( 3x3)
  PROCEDURE near(game_id IN saper_cells.game_id%TYPE,
    cell_x IN saper_cells.x%TYPE, cell_y IN saper_cells.y%TYPE)
  IS
    cell saper_cells%ROWTYPE;
  BEGIN
    cell:=get_cell(cell_x,cell_y,game_id);
    IF NOT(cell.id IS NULL) THEN
      IF cell.state=1 THEN
        FOR op_x IN cell_x-1..cell_x+1
        LOOP
          FOR op_y IN cell_y-1..cell_y+1
          LOOP
            look(game_id,op_x,op_y);
          END LOOP;
        END LOOP;
      END IF;
    END IF;
  END near;
  
END saper;
/
